@@ -24,7 +24,7 @@ module Data.GI.CodeGen.Code
     , recurseWithAPIs
 
     , handleCGExc
-    , describeCGError
+    , printCGError
     , notImplementedError
     , badIntroError
     , missingInfoError
@@ -35,15 +35,20 @@ module Data.GI.CodeGen.Code
     , line
     , blank
     , group
+    , comment
     , cppIf
     , CPPGuard(..)
     , hsBoot
     , submodule
     , setLanguagePragmas
+    , addLanguagePragma
     , setGHCOptions
     , setModuleFlags
     , setModuleMinBase
 
+    , getFreshTypeVariable
+    , resetTypeVariableScope
+
     , exportModule
     , exportDecl
     , export
@@ -71,7 +76,9 @@ import Control.Monad.State.Strict
 import Control.Monad.Except
 import qualified Data.Foldable as F
 import Data.Maybe (fromMaybe, catMaybes)
+#if !MIN_VERSION_base(4,13,0)
 import Data.Monoid ((<>), mempty)
+#endif
 import qualified Data.Map.Strict as M
 import Data.Sequence (ViewL ((:<)), viewl, (|>))
 import qualified Data.Sequence as Seq
@@ -120,6 +127,7 @@ data CodeToken
     = Line Text           -- ^ A single line, indented to current indentation.
     | Indent Code         -- ^ Indented region.
     | Group Code          -- ^ A grouped set of lines
+    | Comment [Text]      -- ^ A (possibly multi line) comment
     | IncreaseIndent      -- ^ Increase the indentation for the rest
                           -- of the lines in the group.
     | CPPBlock CPPConditional Code -- ^ A block of code guarded by the
@@ -233,13 +241,21 @@ data CGError = CGErrorNotImplemented Text
 data CGState = CGState {
   cgsCPPConditionals :: [CPPConditional] -- ^ Active CPP conditionals,
                                          -- outermost condition first.
+  , cgsNextAvailableTyvar :: NamedTyvar -- ^ Next unused type
+                                        -- variable.
   }
 
+-- | The name for a type variable.
+data NamedTyvar = SingleCharTyvar Char
+                -- ^ A single variable type variable: 'a', 'b', etc...
+                | IndexedTyvar Text Integer
+                -- ^ An indexed type variable: 'a17', 'key1', ...
+
 -- | Clean slate for `CGState`.
 emptyCGState :: CGState
-emptyCGState = CGState {
-  cgsCPPConditionals = []
-  }
+emptyCGState = CGState { cgsCPPConditionals = []
+                       , cgsNextAvailableTyvar = SingleCharTyvar 'a'
+                       }
 
 -- | The base type for the code generator monad.
 type BaseCodeGen excType a =
@@ -481,12 +497,14 @@ minBaseVersion minfo =
     maximum (moduleMinBase minfo
             : map minBaseVersion (M.elems $ submodules minfo))
 
--- | Give a friendly textual description of the error for presenting
--- to the user.
-describeCGError :: CGError -> Text
-describeCGError (CGErrorNotImplemented e) = "Not implemented: " <> tshow e
-describeCGError (CGErrorBadIntrospectionInfo e) = "Bad introspection data: " <> tshow e
-describeCGError (CGErrorMissingInfo e) = "Missing info: " <> tshow e
+-- | Print, as a comment, a friendly textual description of the error.
+printCGError :: CGError -> CodeGen ()
+printCGError (CGErrorNotImplemented e) = do
+  comment $ "Not implemented: " <> e
+printCGError (CGErrorBadIntrospectionInfo e) =
+  comment $ "Bad introspection data: " <> e
+printCGError (CGErrorMissingInfo e) =
+  comment $ "Missing info: " <> e
 
 notImplementedError :: Text -> ExcCodeGen a
 notImplementedError s = throwError $ CGErrorNotImplemented s
@@ -497,6 +515,28 @@ badIntroError s = throwError $ CGErrorBadIntrospectionInfo s
 missingInfoError :: Text -> ExcCodeGen a
 missingInfoError s = throwError $ CGErrorMissingInfo s
 
+-- | Get a type variable unused in the current scope.
+getFreshTypeVariable :: CodeGen Text
+getFreshTypeVariable = do
+  (cgs@(CGState{cgsNextAvailableTyvar = available}), s) <- get
+  let (tyvar, next) =
+        case available of
+          SingleCharTyvar char -> case char of
+            'z' -> ("z", IndexedTyvar "a" 0)
+            -- 'm' is reserved for the MonadIO constraint in signatures
+            'm' -> ("n", SingleCharTyvar 'o')
+            c -> (T.singleton c, SingleCharTyvar (toEnum $ fromEnum c + 1))
+          IndexedTyvar root index -> (root <> tshow index,
+                                      IndexedTyvar root (index+1))
+  put (cgs {cgsNextAvailableTyvar = next}, s)
+  return tyvar
+
+-- | Introduce a new scope for type variable naming: the next fresh
+-- variable will be called 'a'.
+resetTypeVariableScope :: CodeGen ()
+resetTypeVariableScope =
+  modify' (\(cgs, s) -> (cgs {cgsNextAvailableTyvar = SingleCharTyvar 'a'}, s))
+
 findAPI :: Type -> CodeGen (Maybe API)
 findAPI TError = Just <$> findAPIByName (Name "GLib" "Error")
 findAPI (TInterface n) = Just <$> findAPIByName n
@@ -535,6 +575,10 @@ bline l = hsBoot (line l) >> line l
 blank :: CodeGen ()
 blank = line ""
 
+-- | A (possibly multi line) comment, separated by newlines
+comment :: Text -> CodeGen ()
+comment = tellCode . Comment . T.lines
+
 -- | Increase the indent level for code generation.
 indent :: BaseCodeGen e a -> BaseCodeGen e a
 indent cg = do
@@ -563,8 +607,8 @@ cppIfBlock cond cg = do
   blank
   return x
     where addConditional :: CGState -> CGState
-          addConditional cgs = CGState {cgsCPPConditionals = CPPIf cond :
-                                         cgsCPPConditionals cgs}
+          addConditional cgs = cgs {cgsCPPConditionals = CPPIf cond :
+                                                         cgsCPPConditionals cgs}
 
 -- | Possible features to test via CPP.
 data CPPGuard = CPPOverloading -- ^ Enable overloading
@@ -572,7 +616,7 @@ data CPPGuard = CPPOverloading -- ^ Enable overloading
 -- | Guard a code block with CPP code, such that it is included only
 -- if the specified feature is enabled.
 cppIf :: CPPGuard -> BaseCodeGen e a -> BaseCodeGen e a
-cppIf CPPOverloading = cppIfBlock "ENABLE_OVERLOADING"
+cppIf CPPOverloading = cppIfBlock "defined(ENABLE_OVERLOADING)"
 
 -- | Write the given code into the .hs-boot file for the current module.
 hsBoot :: BaseCodeGen e a -> BaseCodeGen e a
@@ -609,6 +653,12 @@ setLanguagePragmas :: [Text] -> CodeGen ()
 setLanguagePragmas ps =
     modify' $ \(cgs, s) -> (cgs, s{modulePragmas = Set.fromList ps})
 
+-- | Add a language pragma for the current module.
+addLanguagePragma :: Text -> CodeGen ()
+addLanguagePragma p =
+  modify' $ \(cgs, s) -> (cgs, s{modulePragmas =
+                                 Set.insert p (modulePragmas s)})
+
 -- | Set the GHC options for compiling this module (in a OPTIONS_GHC pragma).
 setGHCOptions :: [Text] -> CodeGen ()
 setGHCOptions opts =
@@ -645,17 +695,29 @@ codeToText (Code seq) = LT.toStrict . B.toLazyText $ genCode 0 (viewl seq)
                                       genCode n (viewl rest)
         genCode n (Group (Code seq) :< rest) = genCode n (viewl seq) <>
                                                genCode n (viewl rest)
+        genCode n (Comment [] :< rest) = genCode n (viewl rest)
+        genCode n (Comment [s] :< rest) =
+          B.fromText (paddedLine n ("-- " <> s)) <> genCode n (viewl rest)
+        genCode n (Comment (l:ls):< rest) =
+          B.fromText ("{-  " <> l <> "\n" <>
+                      paddedLines (n+1) ls <> "-}\n") <> genCode n (viewl rest)
         genCode n (CPPBlock cond (Code seq) :< rest) =
           let (condBegin, condEnd) = cppCondFormat cond
           in B.fromText condBegin <> genCode n (viewl seq) <>
              B.fromText condEnd <> genCode n (viewl rest)
         genCode n (IncreaseIndent :< rest) = genCode (n+1) (viewl rest)
 
--- | Pad a line to the given number of leading spaces, and add a
--- newline at the end.
+-- | Pad a line to the given number of leading tabs (with one tab
+-- equal to four spaces), and add a newline at the end.
 paddedLine :: Int -> Text -> Text
 paddedLine n s = T.replicate (n * 4) " " <> s <> "\n"
 
+-- | Pad a set of lines to the given number of leading tabs (with one
+-- tab equal to four spaces), and add a newline at the end of each
+-- line.
+paddedLines :: Int -> [Text] -> Text
+paddedLines n ls = mconcat $ map (paddedLine n) ls
+
 -- | Put a (padded) comma at the end of the text.
 comma :: Text -> Text
 comma s = padTo 40 s <> ","
@@ -746,7 +808,7 @@ formatSection section exports =
                                                    " #" <> anchor <> "#"
                                     Nothing -> subsectionTitle subsec
                     , case subsectionDoc subsec of
-                        Just text -> "{- | " <> text  <> "\n-}"
+                        Just text -> formatHaddockComment text
                         Nothing -> ""
                     , ( T.concat
                       . map (formatExport exportSymbol)
@@ -789,10 +851,10 @@ ghcOptions opts = "{-# OPTIONS_GHC " <> T.intercalate ", " opts <> " #-}\n"
 
 -- | Generate some convenience CPP macros.
 cppMacros :: Text
-cppMacros = T.unlines ["#define ENABLE_OVERLOADING (MIN_VERSION_haskell_gi_overloading(1,0,0) \\"
-                      -- Haddocks look better without overloading
-                      , "       && !defined(__HADDOCK_VERSION__))"
-                      ]
+cppMacros = T.unlines
+  ["#if (MIN_VERSION_haskell_gi_overloading(1,0,0) && !defined(__HADDOCK_VERSION__))"
+  , "#define ENABLE_OVERLOADING"
+  , "#endif"]
 
 -- | Standard fields for every module.
 standardFields :: Text
@@ -802,9 +864,17 @@ standardFields = T.unlines [ "Copyright  : " <> authors
 
 -- | The haddock header for the module, including optionally a description.
 moduleHaddock :: Maybe Text -> Text
-moduleHaddock Nothing = T.unlines ["{- |", standardFields <> "-}"]
-moduleHaddock (Just description) = T.unlines ["{- |", standardFields,
-                                              description, "-}"]
+moduleHaddock Nothing = formatHaddockComment $ standardFields
+moduleHaddock (Just description) =
+  formatHaddockComment $ T.unlines [standardFields, description]
+
+-- | Format the comment with the module documentation.
+formatHaddockComment :: Text -> Text
+formatHaddockComment doc = let lines = case T.lines doc of
+                                 [] -> []
+                                 (first:rest) -> ("-- | " <> first) :
+                                                 map ("-- " <>) rest
+                          in T.unlines lines
 
 -- | Generic module prelude. We reexport all of the submodules.
 modulePrelude :: M.Map HaddockSection Text -> Text -> [Export] -> [Text] -> Text
@@ -853,15 +923,20 @@ moduleImports = T.unlines [
                 , ""
                 , "import qualified Data.GI.Base.Attributes as GI.Attributes"
                 , "import qualified Data.GI.Base.ManagedPtr as B.ManagedPtr"
+                , "import qualified Data.GI.Base.GClosure as B.GClosure"
                 , "import qualified Data.GI.Base.GError as B.GError"
                 , "import qualified Data.GI.Base.GVariant as B.GVariant"
                 , "import qualified Data.GI.Base.GValue as B.GValue"
                 , "import qualified Data.GI.Base.GParamSpec as B.GParamSpec"
                 , "import qualified Data.GI.Base.CallStack as B.CallStack"
+                , "import qualified Data.GI.Base.Properties as B.Properties"
+                , "import qualified Data.GI.Base.Signals as B.Signals"
+                , "import qualified Control.Monad.IO.Class as MIO"
                 , "import qualified Data.Text as T"
                 , "import qualified Data.ByteString.Char8 as B"
                 , "import qualified Data.Map as Map"
-                , "import qualified Foreign.Ptr as FP" ]
+                , "import qualified Foreign.Ptr as FP"
+                , "import qualified GHC.OverloadedLabels as OL" ]
 
 -- | Like `dotModulePath`, but add a "GI." prefix.
 dotWithPrefix :: ModulePath -> Text
