@@ -4,14 +4,12 @@ module Data.GI.CodeGen.CodeGen
     , genModule
     ) where
 
-#if !MIN_VERSION_base(4,8,0)
-import Control.Applicative ((<$>))
-import Data.Traversable (traverse)
-#endif
 import Control.Monad (forM, forM_, when, unless, filterM)
 import Data.List (nub)
 import Data.Maybe (fromJust, fromMaybe, catMaybes, mapMaybe)
+#if !MIN_VERSION_base(4,13,0)
 import Data.Monoid ((<>))
+#endif
 import qualified Data.Map as M
 import qualified Data.Text as T
 import Data.Text (Text)
@@ -22,7 +20,8 @@ import Data.GI.CodeGen.Constant (genConstant)
 import Data.GI.CodeGen.Code
 import Data.GI.CodeGen.EnumFlags (genEnum, genFlags)
 import Data.GI.CodeGen.Fixups (dropMovedItems, guessPropertyNullability,
-                               detectGObject)
+                               detectGObject, dropDuplicatedFields,
+                               checkClosureDestructors)
 import Data.GI.CodeGen.GObject
 import Data.GI.CodeGen.Haddock (deprecatedPragma, addSectionDocumentation,
                                 writeHaddock,
@@ -39,35 +38,64 @@ import Data.GI.CodeGen.Struct (genStructOrUnionFields, extractCallbacksInStruct,
                   fixAPIStructs, ignoreStruct, genZeroStruct, genZeroUnion,
                   genWrappedPtr)
 import Data.GI.CodeGen.SymbolNaming (upperName, classConstraint, noName,
-                                     submoduleLocation, lowerName)
+                                     submoduleLocation, lowerName, qualifiedAPI)
 import Data.GI.CodeGen.Type
 import Data.GI.CodeGen.Util (tshow)
 
+-- | Standard derived instances for newtypes wrapping @ManagedPtr@s.
+newtypeDeriving :: CodeGen ()
+newtypeDeriving = indent $ line $ "deriving (Eq)"
+
 genFunction :: Name -> Function -> CodeGen ()
 genFunction n (Function symbol fnMovedTo callable) =
     -- Only generate the function if it has not been moved.
     when (Nothing == fnMovedTo) $
       group $ do
-        line $ "-- function " <> symbol
-        handleCGExc (\e -> line ("-- XXX Could not generate function "
-                           <> symbol
-                           <> "\n-- Error was : " <> describeCGError e))
+        line $ "-- function " <> name n
+        handleCGExc (\e -> do
+                        line ("-- XXX Could not generate function "
+                              <> name n
+                              <> "\n")
+                        printCGError e)
                         (do
                           genCCallableWrapper n symbol callable
                           export (NamedSubsection MethodSection $ lowerName n) (lowerName n)
                         )
 
+-- | Generate the GValue instances for the given GObject.
+genBoxedGValueInstance :: Name -> Text -> CodeGen ()
+genBoxedGValueInstance n get_type_fn = do
+  let name' = upperName n
+      doc = "Convert '" <> name' <> "' to and from 'Data.GI.Base.GValue.GValue' with 'Data.GI.Base.GValue.toGValue' and 'Data.GI.Base.GValue.fromGValue'."
+
+  writeHaddock DocBeforeSymbol doc
+
+  group $ do
+    bline $ "instance B.GValue.IsGValue " <> name' <> " where"
+    indent $ group $ do
+      line $ "toGValue o = do"
+      indent $ group $ do
+        line $ "gtype <- " <> get_type_fn
+        line $ "B.ManagedPtr.withManagedPtr o (B.GValue.buildGValue gtype B.GValue.set_boxed)"
+      line $ "fromGValue gv = do"
+      indent $ group $ do
+        line $ "ptr <- B.GValue.get_boxed gv :: IO (Ptr " <> name' <> ")"
+        line $ "B.ManagedPtr.newBoxed " <> name' <> " ptr"
+
 genBoxedObject :: Name -> Text -> CodeGen ()
 genBoxedObject n typeInit = do
   let name' = upperName n
+      get_type_fn = "c_" <> typeInit
 
   group $ do
-    line $ "foreign import ccall \"" <> typeInit <> "\" c_" <>
-            typeInit <> " :: "
+    line $ "foreign import ccall \"" <> typeInit <> "\" " <>
+            get_type_fn <> " :: "
     indent $ line "IO GType"
   group $ do
        line $ "instance BoxedObject " <> name' <> " where"
-       indent $ line $ "boxedType _ = c_" <> typeInit
+       indent $ line $ "boxedType _ = " <> get_type_fn
+
+  genBoxedGValueInstance n get_type_fn
 
   hsBoot $ line $ "instance BoxedObject " <> name' <> " where"
 
@@ -80,6 +108,7 @@ genStruct n s = unless (ignoreStruct n s) $ do
    let decl = line $ "newtype " <> name' <> " = " <> name' <> " (ManagedPtr " <> name' <> ")"
    hsBoot decl
    decl
+   newtypeDeriving
 
    addSectionDocumentation ToplevelSection (structDocumentation s)
 
@@ -103,10 +132,10 @@ genStruct n s = unless (ignoreStruct n s) $ do
        isFunction <- symbolFromFunction (methodSymbol f)
        if not isFunction
        then handleCGExc
-               (\e -> line ("-- XXX Could not generate method "
-                            <> name' <> "::" <> name mn <> "\n"
-                            <> "-- Error was : " <> describeCGError e) >>
-                return Nothing)
+               (\e -> do line ("-- XXX Could not generate method "
+                               <> name' <> "::" <> name mn)
+                         printCGError e
+                         return Nothing)
                (genMethod n f >> return (Just (n, f)))
        else return Nothing
 
@@ -123,6 +152,7 @@ genUnion n u = do
   let decl = line $ "newtype " <> name' <> " = " <> name' <> " (ManagedPtr " <> name' <> ")"
   hsBoot decl
   decl
+  newtypeDeriving
 
   addSectionDocumentation ToplevelSection (unionDocumentation u)
 
@@ -146,10 +176,10 @@ genUnion n u = do
       isFunction <- symbolFromFunction (methodSymbol f)
       if not isFunction
       then handleCGExc
-                (\e -> line ("-- XXX Could not generate method "
-                             <> name' <> "::" <> name mn <> "\n"
-                             <> "-- Error was : " <> describeCGError e)
-                >> return Nothing)
+                (\e -> do line ("-- XXX Could not generate method "
+                                <> name' <> "::" <> name mn)
+                          printCGError e
+                          return Nothing)
                 (genMethod n f >> return (Just (n, f)))
       else return Nothing
 
@@ -234,35 +264,65 @@ genMethod cn m@(Method {
     cppIf CPPOverloading $
          genMethodInfo cn (m {methodCallable = c''})
 
+-- | Generate the GValue instances for the given GObject.
+genGObjectGValueInstance :: Name -> Text -> CodeGen ()
+genGObjectGValueInstance n get_type_fn = do
+  let name' = upperName n
+      doc = "Convert '" <> name' <> "' to and from 'Data.GI.Base.GValue.GValue' with 'Data.GI.Base.GValue.toGValue' and 'Data.GI.Base.GValue.fromGValue'."
+
+  writeHaddock DocBeforeSymbol doc
+
+  group $ do
+    bline $ "instance B.GValue.IsGValue " <> name' <> " where"
+    indent $ group $ do
+      line $ "toGValue o = do"
+      indent $ group $ do
+        line $ "gtype <- " <> get_type_fn
+        line $ "B.ManagedPtr.withManagedPtr o (B.GValue.buildGValue gtype B.GValue.set_object)"
+      line $ "fromGValue gv = do"
+      indent $ group $ do
+        line $ "ptr <- B.GValue.get_object gv :: IO (Ptr " <> name' <> ")"
+        line $ "B.ManagedPtr.newObject " <> name' <> " ptr"
+
 -- Type casting with type checking
 genGObjectCasts :: Name -> Text -> [Name] -> CodeGen ()
 genGObjectCasts n cn_ parents = do
   let name' = upperName n
+      get_type_fn = "c_" <> cn_
 
   group $ do
     line $ "foreign import ccall \"" <> cn_ <> "\""
-    indent $ line $ "c_" <> cn_ <> " :: IO GType"
+    indent $ line $ get_type_fn <> " :: IO GType"
 
   group $ do
     bline $ "instance GObject " <> name' <> " where"
     indent $ group $ do
-            line $ "gobjectType _ = c_" <> cn_
+            line $ "gobjectType = " <> get_type_fn
+
+  genGObjectGValueInstance n get_type_fn
 
   className <- classConstraint n
   group $ do
     exportDecl className
     writeHaddock DocBeforeSymbol (classDoc name')
 
-    bline $ "class GObject o => " <> className <> " o"
-    line $ "#if MIN_VERSION_base(4,9,0)"
-    line $ "instance {-# OVERLAPPABLE #-} (GObject a, O.UnknownAncestorError "
-             <> name' <> " a) =>"
-    line $ "    " <> className <> " a"
-    line $ "#endif"
-    bline $ "instance " <> className <> " " <> name'
-    forM_ parents $ \parent -> do
-        pcls <- classConstraint parent
-        line $ "instance " <> pcls <> " " <> name'
+    -- Create the IsX constraint. We cannot simply say
+    --
+    -- > type IsX o = (GObject o, ...)
+    --
+    -- since we sometimes need to refer to @IsX@ itself, without
+    -- applying it. We instead use the trick of creating a class with
+    -- a universal instance.
+    let constraints = "(GObject o, O.IsDescendantOf " <> name' <> " o)"
+    bline $ "class " <> constraints <> " => " <> className <> " o"
+    bline $ "instance " <> constraints <> " => " <> className <> " o"
+
+    blank
+
+    qualifiedParents <- mapM qualifiedAPI parents
+    bline $ "instance O.HasParentTypes " <> name'
+    line $ "type instance O.ParentTypes " <> name' <> " = '["
+      <> T.intercalate ", " qualifiedParents <> "]"
 
   -- Safe downcasting.
   group $ do
@@ -296,6 +356,7 @@ genObject n o = do
   else do
     writeHaddock DocBeforeSymbol ("Memory-managed wrapper type.")
     bline $ "newtype " <> name' <> " = " <> name' <> " (ManagedPtr " <> name' <> ")"
+    newtypeDeriving
     exportDecl (name' <> "(..)")
 
     addSectionDocumentation ToplevelSection (objDocumentation o)
@@ -309,12 +370,7 @@ genObject n o = do
     cppIf CPPOverloading $
          fullObjectMethodList n o >>= genMethodList n
 
-    forM_ (objSignals o) $ \s ->
-     handleCGExc
-     (line . (T.concat ["-- XXX Could not generate signal ", name', "::"
-                     , sigName s
-                     , "\n", "-- Error was : "] <>) . describeCGError)
-     (genSignal s n)
+    forM_ (objSignals o) $ \s -> genSignal s n
 
     genObjectProperties n o
     cppIf CPPOverloading $
@@ -325,11 +381,11 @@ genObject n o = do
     -- Methods
     forM_ (objMethods o) $ \f -> do
       let mn = methodName f
-      handleCGExc (\e -> line ("-- XXX Could not generate method "
-                              <> name' <> "::" <> name mn <> "\n"
-                              <> "-- Error was : " <> describeCGError e)
-                  >> (cppIf CPPOverloading $
-                           genUnsupportedMethodInfo n f))
+      handleCGExc (\e -> do line ("-- XXX Could not generate method "
+                                  <> name' <> "::" <> name mn)
+                            printCGError e
+                            cppIf CPPOverloading $
+                              genUnsupportedMethodInfo n f)
                   (genMethod n f)
 
 genInterface :: Name -> Interface -> CodeGen ()
@@ -340,6 +396,7 @@ genInterface n iface = do
   writeHaddock DocBeforeSymbol ("Memory-managed wrapper type.")
   deprecatedPragma name' $ ifDeprecated iface
   bline $ "newtype " <> name' <> " = " <> name' <> " (ManagedPtr " <> name' <> ")"
+  newtypeDeriving
   exportDecl (name' <> "(..)")
 
   addSectionDocumentation ToplevelSection (ifDocumentation iface)
@@ -347,9 +404,9 @@ genInterface n iface = do
   noName name'
 
   forM_ (ifSignals iface) $ \s -> handleCGExc
-     (line . (T.concat ["-- XXX Could not generate signal ", name', "::"
-                     , sigName s
-                     , "\n", "-- Error was : "] <>) . describeCGError)
+     (\e -> do line $ T.concat ["-- XXX Could not generate signal ", name', "::"
+                               , sigName s]
+               printCGError e)
      (genSignal s n)
 
   cppIf CPPOverloading $
@@ -374,12 +431,21 @@ genInterface n iface = do
     writeHaddock DocBeforeSymbol ("Type class for types which implement `"
                                   <> name' <> "`.")
 
-    bline $ "class ManagedPtrNewtype a => " <> cls <> " a"
-    line $ "instance " <> cls <> " " <> name'
+    -- Create the IsX constraint. We cannot simply say
+    --
+    -- > type IsX o = (ManagedPtrNewtype o, O.IsDescendantOf X o)
+    --
+    -- since we sometimes need to refer to @IsX@ itself, without
+    -- applying it. We instead use the trick of creating a class with
+    -- a universal instance.
+    let constraints = "(ManagedPtrNewtype o, O.IsDescendantOf " <> name' <> " o)"
+    bline $ "class " <> constraints <> " => " <> cls <> " o"
+    bline $ "instance " <> constraints <> " => " <> cls <> " o"
+
     genWrappedPtr n (ifAllocationInfo iface) 0
 
     when (not . null . ifProperties $ iface) $ group $ do
-       line $ "-- XXX Skipping property generation for non-GObject interface"
+       comment $ "XXX Skipping property generation for non-GObject interface"
 
   -- Methods
   cppIf CPPOverloading $
@@ -390,11 +456,10 @@ genInterface n iface = do
       isFunction <- symbolFromFunction (methodSymbol f)
       unless isFunction $
              handleCGExc
-             (\e -> line ("-- XXX Could not generate method "
-                          <> name' <> "::" <> name mn <> "\n"
-                          <> "-- Error was : " <> describeCGError e)
-             >> (cppIf CPPOverloading $
-                      genUnsupportedMethodInfo n f))
+             (\e -> do comment ("XXX Could not generate method "
+                                <> name' <> "::" <> name mn)
+                       printCGError e
+                       cppIf CPPOverloading (genUnsupportedMethodInfo n f))
              (genMethod n f)
 
 -- Some type libraries include spurious interface/struct methods,
@@ -449,6 +514,13 @@ genModule' apis = do
             -- Not every interface providing signals or properties is
             -- correctly annotated as descending from GObject, fix this.
           $ map detectGObject
+            -- Some APIs contain duplicated fields by mistake, drop
+            -- the duplicates.
+          $ map dropDuplicatedFields
+            -- Make sure that every argument marked as being a
+            -- destructor for a user_data argument has an associated
+            -- user_data argument.
+          $ map checkClosureDestructors
           $ M.toList
           $ apis
 
