@@ -16,6 +16,7 @@ module Data.GI.CodeGen.Conversions
     , foreignType
 
     , argumentType
+    , ExposeClosures(..)
     , elementType
     , elementMap
     , elementTypeAndMap
@@ -41,7 +42,9 @@ import Control.Applicative ((<$>), (<*>), pure, Applicative)
 #endif
 import Control.Monad (when)
 import Data.Maybe (isJust)
+#if !MIN_VERSION_base(4,11,0)
 import Data.Monoid ((<>))
+#endif
 import Data.Text (Text)
 import qualified Data.Text as T
 import GHC.Exts (IsString(..))
@@ -212,6 +215,21 @@ hParamSpecToF transfer =
   then return $ M "B.GParamSpec.disownGParamSpec"
   else return $ M "unsafeManagedPtrGetPtr"
 
+hClosureToF :: Transfer -> Maybe Type -> CodeGen Constructor
+-- Untyped closures
+hClosureToF transfer Nothing =
+  if transfer == TransferEverything
+  then return $ M "B.GClosure.disownGClosure"
+  -- We cast the point here because the foreign type for untyped
+  -- closures is always represented as Ptr (GClosure ()), while the
+  -- corresponding Haskell type is the parametric "GClosure a".
+  else return $ M "unsafeManagedPtrCastPtr"
+-- Typed closures
+hClosureToF transfer (Just _) =
+  if transfer == TransferEverything
+  then return $ M "B.GClosure.disownGClosure"
+  else return $ M "unsafeManagedPtrGetPtr"
+
 hBoxedToF :: Transfer -> CodeGen Constructor
 hBoxedToF transfer =
   if transfer == TransferEverything
@@ -245,6 +263,7 @@ hToF' t a hType fType transfer
     | TError <- t = hBoxedToF transfer
     | TVariant <- t = hVariantToF transfer
     | TParamSpec <- t = hParamSpecToF transfer
+    | TGClosure c <- t = hClosureToF transfer c
     | Just (APIEnum _) <- a = return "(fromIntegral . fromEnum)"
     | Just (APIFlags _) <- a = return "gflagsToWord"
     | Just (APIObject _) <- a = hObjectToF t transfer
@@ -460,6 +479,19 @@ fParamSpecToH transfer =
                   TransferEverything -> "B.GParamSpec.wrapGParamSpecPtr"
                   _ -> "B.GParamSpec.newGParamSpecFromPtr"
 
+fClosureToH :: Transfer -> Maybe Type -> CodeGen Constructor
+-- Untyped closures
+fClosureToH transfer Nothing =
+  return $ M $ case transfer of
+                  TransferEverything ->
+                    parenthesize $ "B.GClosure.wrapGClosurePtr . FP.castPtr"
+                  _ -> parenthesize $ "B.GClosure.newGClosureFromPtr . FP.castPtr"
+-- Typed closures
+fClosureToH transfer (Just _) =
+  return $ M $ case transfer of
+                  TransferEverything -> "B.GClosure.wrapGClosurePtr"
+                  _ -> "B.GClosure.newGClosureFromPtr"
+
 fToH' :: Type -> Maybe API -> TypeRep -> TypeRep -> Transfer
          -> ExcCodeGen Constructor
 fToH' t a hType fType transfer
@@ -469,6 +501,7 @@ fToH' t a hType fType transfer
     | TError <- t = boxedForeignPtr "GError" transfer
     | TVariant <- t = fVariantToH transfer
     | TParamSpec <- t = fParamSpecToH transfer
+    | TGClosure c <- t = fClosureToH transfer c
     | Just (APIStruct s) <- a = structForeignPtr s hType transfer
     | Just (APIUnion u) <- a = unionForeignPtr u hType transfer
     | Just (APIObject _) <- a = fObjectToH t hType transfer
@@ -662,18 +695,23 @@ unpackCArray length (TCArray False _ _ t) transfer =
 
 unpackCArray _ _ _ = notImplementedError "unpackCArray : unexpected array type."
 
+-- | Whether to expose closures and the associated destroy notify
+-- handlers in the Haskell wrapper.
+data ExposeClosures = WithClosures
+                    | WithoutClosures
+  deriving (Eq)
+
 -- | Given a type find the typeclasses the type belongs to, and return
 -- the representation of the type in the function signature and the
 -- list of typeclass constraints for the type.
-argumentType :: [Char] -> Type -> CodeGen ([Char], Text, [Text])
-argumentType [] _               = error "out of letters"
-argumentType letters (TGList a) = do
-  (ls, name, constraints) <- argumentType letters a
-  return (ls, "[" <> name <> "]", constraints)
-argumentType letters (TGSList a) = do
-  (ls, name, constraints) <- argumentType letters a
-  return (ls, "[" <> name <> "]", constraints)
-argumentType letters@(l:ls) t = do
+argumentType :: Type -> ExposeClosures -> CodeGen (Text, [Text])
+argumentType (TGList a) expose = do
+  (name, constraints) <- argumentType a expose
+  return ("[" <> name <> "]", constraints)
+argumentType (TGSList a) expose = do
+  (name, constraints) <- argumentType a expose
+  return ("[" <> name <> "]", constraints)
+argumentType t expose = do
   api <- findAPI t
   s <- typeShow <$> haskellType t
   case api of
@@ -681,22 +719,29 @@ argumentType letters@(l:ls) t = do
     -- we allow for any object descending from it.
     Just (APIInterface _) -> do
       cls <- typeConstraint t
-      return (ls, T.singleton l, [cls <> " " <> T.singleton l])
+      l <- getFreshTypeVariable
+      return (l, [cls <> " " <> l])
     Just (APIObject _) -> do
       isGO <- isGObject t
       if isGO
         then do cls <- typeConstraint t
-                return (ls, T.singleton l, [cls <> " " <> T.singleton l])
-        else return (letters, s, [])
+                l <- getFreshTypeVariable
+                return (l, [cls <> " " <> l])
+        else return (s, [])
     Just (APICallback cb) ->
       -- See [Note: Callables that throw]
       if callableThrows (cbCallable cb)
       then do
         ft <- typeShow <$> foreignType t
-        return (letters, ft, [])
+        return (ft, [])
       else
-        return (letters, s, [])
-    _ -> return (letters, s, [])
+        case expose of
+          WithClosures -> do
+            s_withClosures <- typeShow <$> isoHaskellType t
+            return (s_withClosures, [])
+          WithoutClosures ->
+            return (s, [])
+    _ -> return (s, [])
 
 haskellBasicType :: BasicType -> TypeRep
 haskellBasicType TPtr      = ptr $ con0 "()"
@@ -763,7 +808,18 @@ haskellType (TGHash a b) = do
 haskellType TError = return $ "GError" `con` []
 haskellType TVariant = return $ "GVariant" `con` []
 haskellType TParamSpec = return $ "GParamSpec" `con` []
-haskellType (TInterface (Name "GObject" "Closure")) = return $ "Closure" `con` []
+haskellType (TGClosure (Just inner@(TInterface n))) = do
+  innerAPI <- getAPI inner
+  case innerAPI of
+    APICallback _ -> do
+      tname <- qualifiedSymbol (callbackCType $ name n) n
+      return $ "GClosure" `con` [con0 tname]
+    -- The given inner type does not make sense, so we treat it as an
+    -- untyped closure.
+    _ -> haskellType (TGClosure Nothing)
+haskellType (TGClosure _) = do
+  tyvar <- getFreshTypeVariable
+  return $ "GClosure" `con` [con0 tyvar]
 haskellType (TInterface (Name "GObject" "Value")) = return $ "GValue" `con` []
 haskellType t@(TInterface n) = do
   api <- getAPI t
@@ -786,13 +842,21 @@ typeIsCallback t@(TInterface _) = do
     _ -> return False
 typeIsCallback _ = return False
 
--- | Basically like `haskellType`, but for types which admit a "isomorphic"
--- version of the Haskell type distinct from the usual Haskell type.
--- Generally the Haskell type we expose is isomorphic to the foreign
--- type, but in some cases, such as callbacks with closure arguments,
--- this does not hold, as we omit the closure arguments. This function
--- returns a type which is actually isomorphic.
+-- | Basically like `haskellType`, but for types which admit a
+-- "isomorphic" version of the Haskell type distinct from the usual
+-- Haskell type.  Generally the Haskell type we expose is isomorphic
+-- to the foreign type, but in some cases, such as callbacks with
+-- closure arguments, this does not hold, as we omit the closure
+-- arguments. This function returns a type which is actually
+-- isomorphic. There is another case this function deals with: for
+-- convenience untyped `TGClosure` types have a type variable on the
+-- Haskell side when they are arguments to functions, but we do not
+-- want this when they appear as arguments to callbacks/signals, or
+-- return types of properties, as it would force the type synonym/type
+-- family to depend on the type variable.
 isoHaskellType :: Type -> CodeGen TypeRep
+isoHaskellType (TGClosure Nothing) =
+  return $ "GClosure" `con` [con0 "()"]
 isoHaskellType t@(TInterface n) = do
   api <- findAPI t
   case api of
@@ -847,8 +911,8 @@ foreignType (TGHash a b) = do
 foreignType t@TError = ptr <$> haskellType t
 foreignType t@TVariant = ptr <$> haskellType t
 foreignType t@TParamSpec = ptr <$> haskellType t
-foreignType (TInterface (Name "GObject" "Closure")) =
-    return $ ptr $ "Closure" `con` []
+foreignType (TGClosure Nothing) = return $ ptr ("GClosure" `con` [con0 "()"])
+foreignType t@(TGClosure (Just _)) = ptr <$> haskellType t
 foreignType (TInterface (Name "GObject" "Value")) =
   return $ ptr $ "GValue" `con` []
 foreignType t@(TInterface n) = do
@@ -903,6 +967,7 @@ isManaged   :: Type -> CodeGen Bool
 isManaged TError = return True
 isManaged TVariant = return True
 isManaged TParamSpec = return True
+isManaged (TGClosure _) = return True
 isManaged t@(TInterface _) = do
   a <- findAPI t
   case a of
