@@ -6,31 +6,37 @@ module Data.GI.CodeGen.CtoHaskellMap
   ) where
 
 import qualified Data.Map as M
+import Data.Maybe (catMaybes)
+#if !MIN_VERSION_base(4,11,0)
 import Data.Monoid ((<>))
+#endif
 import Data.Text (Text)
-import qualified Data.Text as T
-import Data.String (IsString(..))
 
 import Data.GI.CodeGen.GtkDoc (CRef(..))
 import Data.GI.CodeGen.API (API(..), Name(..), Callback(..),
                             Constant(..), Flags(..),
                             Enumeration(..), EnumerationMember(..),
                             Interface(..), Object(..),
-                            Function(..), Method(..), Struct(..), Union(..))
+                            Function(..), Method(..), Struct(..), Union(..),
+                            Signal(..))
 import Data.GI.CodeGen.ModulePath (ModulePath, dotModulePath, (/.))
-import Data.GI.CodeGen.SymbolNaming (submoduleLocation, lowerName, upperName)
+import Data.GI.CodeGen.SymbolNaming (submoduleLocation, lowerName, upperName,
+                                     signalHaskellName)
 import Data.GI.CodeGen.Util (ucFirst)
 
 -- | Link to an identifier, module, etc.
-data Hyperlink = IdentifierLink Text
+data Hyperlink = ValueIdentifier Text
+               -- ^ An identifier at the value level: functions, data
+               -- constructors, ...
+               | TypeIdentifier Text
+               -- ^ An identifier at the type level.
                | ModuleLink Text
-               | ModuleLinkWithAnchor Text Text
+               -- ^ Link to a module.
+               | ModuleLinkWithAnchor (Maybe Text) Text Text
+               -- ^ Link to an anchor inside a given module, with an
+               -- optional label.
   deriving (Show, Eq)
 
--- Just for convenience
-instance IsString Hyperlink where
-  fromString = IdentifierLink . T.pack
-
 -- | Given a set of APIs, build a `Map` that given a Text
 -- corresponding to a certain C identifier returns the corresponding
 -- Haskell element in the bindings. For instance, `gtk_widget_show`
@@ -50,63 +56,83 @@ cToHaskellMap apis = M.union (M.fromList builtins)
         extractRefs (n, APIObject o) = objectRefs n o
 
         builtins :: [(CRef, Hyperlink)]
-        builtins = [(TypeRef "gboolean", "Bool"),
-                    (ConstantRef "TRUE", "True"),
-                    (ConstantRef "FALSE", "False"),
-                    (TypeRef "GError", "GError"),
-                    (TypeRef "GType", "GType"),
-                    (TypeRef "GVariant", "GVariant"),
-                    (ConstantRef "NULL", "Nothing")]
+        builtins = [(TypeRef "gboolean", TypeIdentifier "P.Bool"),
+                    (ConstantRef "TRUE", ValueIdentifier "P.True"),
+                    (ConstantRef "FALSE", ValueIdentifier "P.False"),
+                    (TypeRef "GError", TypeIdentifier "GError"),
+                    (TypeRef "GType", TypeIdentifier "GType"),
+                    (TypeRef "GVariant", TypeIdentifier "GVariant"),
+                    (ConstantRef "NULL", ValueIdentifier "P.Nothing")]
 
 -- | Obtain the absolute location of the module where the given `API`
 -- lives.
 location :: Name -> API -> ModulePath
 location n api = ("GI" /. ucFirst (namespace n)) <> submoduleLocation n api
 
--- | Obtain the fully qualified symbol.
-fullyQualified :: Name -> API -> Text -> Hyperlink
-fullyQualified n api symbol =
-  IdentifierLink $ dotModulePath (location n api) <> "." <> symbol
+-- | Obtain the fully qualified symbol pointing to a value.
+fullyQualifiedValue :: Name -> API -> Text -> Hyperlink
+fullyQualifiedValue n api symbol =
+  ValueIdentifier $ dotModulePath (location n api) <> "." <> symbol
+
+-- | Obtain the fully qualified symbol pointing to a type.
+fullyQualifiedType :: Name -> API -> Text -> Hyperlink
+fullyQualifiedType n api symbol =
+  TypeIdentifier $ dotModulePath (location n api) <> "." <> symbol
 
 -- | Extract the C name of a constant. These are often referred to as
 -- types, so we allow that too.
 constRefs :: Name -> Constant -> [(CRef, Hyperlink)]
 constRefs n c = [(ConstantRef (constantCType c),
-                  fullyQualified n (APIConst c) $ name n),
+                  fullyQualifiedValue n (APIConst c) $ name n),
                  (TypeRef (constantCType c),
-                  fullyQualified n (APIConst c) $ name n)]
+                  fullyQualifiedValue n (APIConst c) $ name n)]
 
 -- | Extract the C name of a function.
 funcRefs :: Name -> Function -> [(CRef, Hyperlink)]
 funcRefs n f = [(FunctionRef (fnSymbol f),
-                 fullyQualified n (APIFunction f) $ lowerName n)]
+                 fullyQualifiedValue n (APIFunction f) $ lowerName n)]
 
 -- | Extract the C names of the fields in an enumeration/flags, and
 -- the name of the type itself.
 enumRefs :: API -> Name -> Enumeration -> [(CRef, Hyperlink)]
-enumRefs api n e = (TypeRef (enumCType e), fullyQualified n api $ upperName n) :
+enumRefs api n e = (TypeRef (enumCType e),
+                    fullyQualifiedType n api $ upperName n) :
                    map memberToRef (enumMembers e)
   where memberToRef :: EnumerationMember -> (CRef, Hyperlink)
         memberToRef em = (ConstantRef (enumMemberCId em),
-                          fullyQualified n api $ upperName $
+                          fullyQualifiedValue n api $ upperName $
                           n {name = name n <> "_" <> enumMemberName em})
 
+-- | Refs to the methods for a given owner.
+methodRefs :: Name -> API -> [Method] -> [(CRef, Hyperlink)]
+methodRefs n api methods = catMaybes $ map methodRef methods
+  where methodRef :: Method -> Maybe (CRef, Hyperlink)
+        methodRef Method{methodSymbol = symbol, methodName = mn} =
+          -- Method name namespaced by the owner.
+          let mn' = mn {name = name n <> "_" <> name mn}
+          in Just (FunctionRef symbol,
+                   fullyQualifiedValue n api $ lowerName mn')
+
+-- | Refs to the signals for a given owner.
+signalRefs :: Name -> API -> Maybe Text -> [Signal] -> [(CRef, Hyperlink)]
+signalRefs n api maybeCName signals = map signalRef signals
+  where signalRef :: Signal -> (CRef, Hyperlink)
+        signalRef (Signal {sigName = sn}) =
+          let mod = dotModulePath (location n api)
+              sn' = signalHaskellName sn
+              ownerCName = case maybeCName of
+                Just cname -> cname
+                Nothing -> let Name ns owner = n
+                           in ucFirst ns <> owner
+          in (SignalRef ownerCName sn,
+              ModuleLinkWithAnchor (Just sn') mod ("signal:" <> sn'))
+
 -- | Given an optional C type and the API constructor construct the
 -- list of associated refs.
 maybeCType :: Name -> API -> Maybe Text -> [(CRef, Hyperlink)]
 maybeCType _ _ Nothing = []
 maybeCType n api (Just ctype) = [(TypeRef ctype,
-                                  fullyQualified n api (upperName n))]
-
--- | Refs to the methods for a given owner.
-methodRefs :: Name -> API -> [Method] -> [(CRef, Hyperlink)]
-methodRefs n api methods = map methodRef methods
-  where methodRef :: Method -> (CRef, Hyperlink)
-        methodRef m@(Method {methodName = mn}) =
-          -- Method name namespaced by the owner.
-          let mn' = mn {name = name n <> "_" <> name mn}
-          in (FunctionRef (methodSymbol m),
-              fullyQualified n api $ lowerName mn')
+                                  fullyQualifiedType n api (upperName n))]
 
 -- | Extract the C name of a callback.
 callbackRefs :: Name -> Callback -> [(CRef, Hyperlink)]
@@ -126,8 +152,10 @@ unionRefs n u = maybeCType n (APIUnion u) (unionCType u)
 ifaceRefs :: Name -> Interface -> [(CRef, Hyperlink)]
 ifaceRefs n i = maybeCType n (APIInterface i) (ifCType i)
                  <> methodRefs n (APIInterface i) (ifMethods i)
+                 <> signalRefs n (APIInterface i) (ifCType i) (ifSignals i)
 
 -- | Extract the C references in an object.
 objectRefs :: Name -> Object -> [(CRef, Hyperlink)]
 objectRefs n o = maybeCType n (APIObject o) (objCType o)
                  <> methodRefs n (APIObject o) (objMethods o)
+                 <> signalRefs n (APIObject o) (objCType o) (objSignals o)
