@@ -10,12 +10,15 @@ module Data.GI.CodeGen.Haddock
   , addSectionDocumentation
   ) where
 
-#if !MIN_VERSION_base(4,8,0)
-import Control.Applicative ((<$>))
-#endif
+#if !MIN_VERSION_base(4,13,0)
 import Control.Monad (mapM_, unless)
+#else
+import Control.Monad (unless)
+#endif
 import qualified Data.Map as M
+#if !MIN_VERSION_base(4,11,0)
 import Data.Monoid ((<>))
+#endif
 import qualified Data.Text as T
 import Data.Text (Text)
 
@@ -31,7 +34,7 @@ import Data.GI.CodeGen.CtoHaskellMap (Hyperlink(..))
 import Data.GI.CodeGen.GtkDoc (GtkDoc(..), Token(..), CRef(..), Language(..),
                                Link(..), ListItem(..), parseGtkDoc)
 import Data.GI.CodeGen.Overrides (onlineDocsMap)
-import Data.GI.CodeGen.SymbolNaming (lowerSymbol)
+import Data.GI.CodeGen.SymbolNaming (lowerSymbol, signalHaskellName)
 
 -- | Where is the documentation located with respect to the relevant
 -- symbol, useful for determining whether we want to start with @|@ or @^@.
@@ -47,9 +50,9 @@ data RelativeDocPosition = DocBeforeSymbol
 -- >>> formatHaddock M.empty "" (GtkDoc [Literal "Hello ", Literal "World!"])
 -- "Hello World!"
 --
--- >>> let c2h = M.fromList [(FunctionRef "foo", "foo()")]
+-- >>> let c2h = M.fromList [(FunctionRef "foo", ValueIdentifier "foo")]
 -- >>> formatHaddock c2h "" (GtkDoc [SymbolRef (FunctionRef "foo")])
--- "'foo()'"
+-- "'foo'"
 --
 -- >>> let onlineDocs = "http://wiki.haskell.org"
 -- >>> formatHaddock M.empty onlineDocs (GtkDoc [ExternalLink (Link "GI" "GObjectIntrospection")])
@@ -61,13 +64,13 @@ formatHaddock :: M.Map CRef Hyperlink -> Text -> GtkDoc -> Text
 formatHaddock c2h docBase (GtkDoc doc) = T.concat $ map formatToken doc
   where formatToken :: Token -> Text
         formatToken (Literal l) = escape l
+        formatToken (Comment _) = ""
         formatToken (Verbatim v) = "@" <> escape v <> "@"
         formatToken (CodeBlock l c) = formatCodeBlock l c
         formatToken (ExternalLink l) = formatLink l docBase
         formatToken (Image l) = formatImage l docBase
         formatToken (SectionHeader l h) = formatSectionHeader c2h docBase l h
         formatToken (List l) = formatList c2h docBase l
-        formatToken (SymbolRef (ParamRef p)) = "/@" <> lowerSymbol p <> "@/"
         formatToken (SymbolRef cr) = case M.lookup cr c2h of
           Just hr -> formatHyperlink hr
           Nothing -> formatUnknownCRef c2h cr
@@ -75,7 +78,10 @@ formatHaddock c2h docBase (GtkDoc doc) = T.concat $ map formatToken doc
 -- | Format a `CRef` whose Haskell representation is not known.
 formatUnknownCRef :: M.Map CRef Hyperlink -> CRef -> Text
 formatUnknownCRef _ (FunctionRef f) = formatCRef $ f <> "()"
-formatUnknownCRef _ (ParamRef _) = error $ "Should not be reached"
+formatUnknownCRef _ (ParamRef p) = "/@" <> lowerSymbol p <> "@/"
+formatUnknownCRef _ (LocalSignalRef s) =
+  let sn = signalHaskellName s
+  in "[" <> sn <> "](#signal:" <> sn <> ")"
 formatUnknownCRef c2h (SignalRef owner signal) =
   case M.lookup (TypeRef owner) c2h of
     Nothing -> formatCRef $ owner <> "::" <> signal
@@ -101,9 +107,13 @@ formatCRef t = "@/" <> escape t <> "/@"
 
 -- | Format a `Hyperlink` into plain `Text`.
 formatHyperlink :: Hyperlink -> Text
-formatHyperlink (IdentifierLink t) = "'" <> t <> "'"
+formatHyperlink (TypeIdentifier t) = "t'" <> t <> "'"
+formatHyperlink (ValueIdentifier t) = "'" <> t <> "'"
 formatHyperlink (ModuleLink m) = "\"" <> m <> "\""
-formatHyperlink (ModuleLinkWithAnchor m a) = "\"" <> m <> "#" <> a <> "\""
+formatHyperlink (ModuleLinkWithAnchor mLabel m a) =
+  case mLabel of
+    Nothing -> "\"" <> m <> "#" <> a <> "\""
+    Just label -> "[" <> label <> "](\"" <> m <> "#" <> a <> "\")"
 
 -- | Format a code block in a specified language.
 formatCodeBlock :: Maybe Language -> Text -> Text
@@ -220,14 +230,9 @@ formatDocumentation c2h docBase doc = do
 -- | Write the given documentation into generated code.
 writeDocumentation :: RelativeDocPosition -> Documentation -> CodeGen ()
 writeDocumentation pos doc = do
-  line $ case pos of
-           DocBeforeSymbol -> "{- |"
-           DocAfterSymbol ->  "{- ^"
   c2h <- getC2HMap
   docBase <- getDocBase
-  let haddock = formatDocumentation c2h docBase doc
-  mapM_ line (T.lines haddock)
-  line "-}"
+  writeHaddock pos (formatDocumentation c2h docBase doc)
 
 -- | Like `writeDocumentation`, but allows us to pass explicitly the
 -- Haddock comment to write.
@@ -236,12 +241,10 @@ writeHaddock pos haddock =
   let marker = case pos of
         DocBeforeSymbol -> "|"
         DocAfterSymbol -> "^"
-  in if T.any (== '\n') haddock
-     then do
-        line $ "{- " <> marker
-        mapM_ line (T.lines haddock)
-        line $ "-}"
-     else line $ "-- " <> marker <> " " <> haddock
+      lines = case T.lines haddock of
+        [] -> []
+        (first:rest) -> ("-- " <> marker <> " " <> first) : map ("-- " <>) rest
+  in mapM_ line lines
 
 -- | Write the documentation for the given argument.
 writeArgDocumentation :: Arg -> CodeGen ()
@@ -251,8 +254,9 @@ writeArgDocumentation arg =
     Just raw -> do
       c2h <- getC2HMap
       docBase <- getDocBase
-      line $ "{- ^ /@" <> lowerSymbol (argCName arg) <> "@/: " <>
-        formatHaddock c2h docBase (parseGtkDoc raw) <> " -}"
+      let haddock = "/@" <> lowerSymbol (argCName arg) <> "@/: " <>
+                    formatHaddock c2h docBase (parseGtkDoc raw)
+      writeHaddock DocAfterSymbol haddock
 
 -- | Write the documentation for the given return value.
 writeReturnDocumentation :: Callable -> Bool -> CodeGen ()
@@ -271,7 +275,7 @@ writeReturnDocumentation callable skip = do
                    else []
   let fullInfo = T.intercalate " " (returnValInfo ++ throwsInfo)
   unless (T.null fullInfo) $
-    line $ "{- ^ " <>  fullInfo <> " -}"
+    writeHaddock DocAfterSymbol fullInfo
 
 -- | Add the given text to the documentation for the section being generated.
 addSectionDocumentation :: HaddockSection -> Documentation -> CodeGen ()
