@@ -8,7 +8,9 @@ module Data.GI.CodeGen.Properties
 import Control.Applicative ((<$>))
 #endif
 import Control.Monad (forM_, when, unless)
+#if !MIN_VERSION_base(4,11,0)
 import Data.Monoid ((<>))
+#endif
 import Data.Text (Text)
 import qualified Data.Text as T
 import qualified Data.Set as S
@@ -23,10 +25,11 @@ import Data.GI.CodeGen.GObject
 import Data.GI.CodeGen.Haddock (addSectionDocumentation, writeHaddock,
                                 RelativeDocPosition(DocBeforeSymbol))
 import Data.GI.CodeGen.Inheritance (fullObjectPropertyList, fullInterfacePropertyList)
-import Data.GI.CodeGen.SymbolNaming (lowerName, upperName,
-                                     classConstraint, typeConstraint,
+import Data.GI.CodeGen.SymbolNaming (lowerName, upperName, classConstraint,
                                      hyphensToCamelCase, qualifiedSymbol,
-                                     callbackDynamicWrapper)
+                                     typeConstraint, callbackDynamicWrapper,
+                                     callbackHaskellToForeign,
+                                     callbackWrapperAllocator)
 import Data.GI.CodeGen.Type
 import Data.GI.CodeGen.Util
 
@@ -39,6 +42,8 @@ propTypeStr t = case t of
    TGHash _ _ -> return "Hash"
    TVariant -> return "Variant"
    TParamSpec -> return "ParamSpec"
+   TGClosure _ -> return "Closure"
+   TError -> return "GError"
    TBasicType TInt -> case sizeOf (0 :: CInt) of
                         4 -> return "Int32"
                         n -> error ("Unsupported `gint' type length: " ++
@@ -85,17 +90,82 @@ propTypeStr t = case t of
        _ -> error $ "Unknown interface property of type : " ++ show t
    _ -> error $ "Don't know how to handle properties of type " ++ show t
 
+-- | The constraint for setting the given type in properties.
+propSetTypeConstraint :: Type -> CodeGen Text
+propSetTypeConstraint (TGClosure Nothing) =
+  return $ "(~) " <> parenthesize (typeShow ("GClosure" `con` [con0 "()"]))
+propSetTypeConstraint t = do
+  isGO <- isGObject t
+  if isGO
+    then typeConstraint t
+    else do
+      isCallback <- typeIsCallback t
+      hInType <- if isCallback
+                 then typeShow <$> foreignType t
+                 else typeShow <$> haskellType t
+      return $ "(~) " <> if T.any (== ' ') hInType
+                         then parenthesize hInType
+                         else hInType
+
+-- | The constraint for transferring the given type into a property.
+propTransferTypeConstraint :: Type -> CodeGen Text
+propTransferTypeConstraint t = do
+  isGO <- isGObject t
+  if isGO
+    then typeConstraint t
+    else do
+      hInType <- typeShow <$> isoHaskellType t
+      return $ "(~) " <> if T.any (== ' ') hInType
+                         then parenthesize hInType
+                         else hInType
+
+-- | The type of the return value of @attrTransfer@ for the given
+-- type.
+propTransferType :: Type -> CodeGen Text
+propTransferType (TGClosure Nothing) =
+  return $ typeShow ("GClosure" `con` [con0 "()"])
+propTransferType t = do
+  isCallback <- typeIsCallback t
+  if isCallback
+             then typeShow <$> foreignType t
+             else typeShow <$> haskellType t
+
+-- | Given a value "v" of the given Haskell type, satisfying the
+-- constraint generated by 'propTransferTypeConstraint', convert it
+-- (allocating memory is necessary) to the type given by 'propTransferType'.
+genPropTransfer :: Text -> Type -> CodeGen ()
+genPropTransfer var (TGClosure Nothing) = line $ "return " <> var
+genPropTransfer var t = do
+  isGO <- isGObject t
+  if isGO
+    then do
+      ht <- typeShow <$> haskellType t
+      line $ "unsafeCastTo " <> ht <> " " <> var
+    else case t of
+           TInterface tn@(Name _ n) -> do
+             isCallback <- typeIsCallback t
+             if not isCallback
+               then line $ "return " <> var
+               else do
+               -- Callbacks need to be wrapped
+               wrapper <- qualifiedSymbol (callbackHaskellToForeign n) tn
+               maker <- qualifiedSymbol (callbackWrapperAllocator n) tn
+               line $ maker <> " " <>
+                 parenthesize (wrapper <> " Nothing " <> var)
+           _ -> line $ "return " <> var
+
 -- | Given a property, return the set of constraints on the types, and
 -- the type variables for the object and its value.
 attrType :: Property -> CodeGen ([Text], Text)
 attrType prop = do
+  resetTypeVariableScope
   isCallback <- typeIsCallback (propType prop)
   if isCallback
     then do
       ftype <- foreignType (propType prop)
       return ([], typeShow ftype)
     else do
-      (_,t,constraints) <- argumentType ['a'..'l'] $ propType prop
+      (t,constraints) <- argumentType (propType prop) WithoutClosures
       return (constraints, t)
 
 -- | Generate documentation for the given setter.
@@ -120,7 +190,7 @@ genPropertySetter setter n docSection prop = group $ do
   writeHaddock DocBeforeSymbol (setterDoc n prop)
   line $ setter <> " :: (" <> T.intercalate ", " constraints'
            <> ") => o -> " <> t <> " -> m ()"
-  line $ setter <> " obj val = liftIO $ setObjectProperty" <> tStr
+  line $ setter <> " obj val = liftIO $ B.Properties.setObjectProperty" <> tStr
            <> " obj \"" <> propName prop
            <> if isNullable && (not isCallback)
               then "\" (Just val)"
@@ -151,8 +221,8 @@ genPropertyGetter getter n docSection prop = group $ do
       returnType = typeShow $ "m" `con` [outType]
       getProp = if isNullable && not isMaybe
                 then "checkUnexpectedNothing \"" <> getter
-                         <> "\" $ getObjectProperty" <> tStr
-                else "getObjectProperty" <> tStr
+                         <> "\" $ B.Properties.getObjectProperty" <> tStr
+                else "B.Properties.getObjectProperty" <> tStr
   -- Some property getters require in addition a constructor, which
   -- will convert the foreign value to the wrapped Haskell one.
   constructorArg <-
@@ -184,15 +254,15 @@ genPropertyConstructor constructor n docSection prop = group $ do
   isNullable <- typeIsNullable (propType prop)
   isCallback <- typeIsCallback (propType prop)
   cls <- classConstraint n
-  let constraints' = (cls <> " o") : constraints
+  let constraints' = (cls <> " o") : "MIO.MonadIO m" : constraints
       pconstraints = parenthesize (T.intercalate ", " constraints') <> " => "
   writeHaddock DocBeforeSymbol (constructorDoc prop)
   line $ constructor <> " :: " <> pconstraints
-           <> t <> " -> IO (GValueConstruct o)"
-  line $ constructor <> " val = constructObjectProperty" <> tStr
+           <> t <> " -> m (GValueConstruct o)"
+  line $ constructor <> " val = MIO.liftIO $ B.Properties.constructObjectProperty" <> tStr
            <> " \"" <> propName prop
            <> if isNullable && (not isCallback)
-              then "\" (Just val)"
+              then "\" (P.Just val)"
               else "\" val"
   export docSection constructor
 
@@ -219,7 +289,7 @@ genPropertyClear clear n docSection prop = group $ do
                 else "(Nothing :: " <> nothingType <> ")"
   line $ clear <> " :: (" <> T.intercalate ", " constraints
            <> ") => o -> m ()"
-  line $ clear <> " obj = liftIO $ setObjectProperty" <> tStr
+  line $ clear <> " obj = liftIO $ B.Properties.setObjectProperty" <> tStr
            <> " obj \"" <> propName prop <> "\" " <> nothing
   export docSection clear
 
@@ -332,18 +402,14 @@ genOneProperty owner prop = do
   cppIf CPPOverloading $ do
     cls <- classConstraint owner
     inConstraint <- if writable || constructOnly
-                    then do
-                      inIsGO <- isGObject (propType prop)
-                      isCallback <- typeIsCallback (propType prop)
-                      hInType <- if isCallback
-                                 then typeShow <$> foreignType (propType prop)
-                                 else typeShow <$> haskellType (propType prop)
-                      if inIsGO
-                         then typeConstraint (propType prop)
-                         else return $ "(~) " <> if T.any (== ' ') hInType
-                                                 then parenthesize hInType
-                                                 else hInType
+                    then propSetTypeConstraint (propType prop)
                     else return "(~) ()"
+    transferConstraint <- if writable || constructOnly
+                          then propTransferTypeConstraint (propType prop)
+                          else return "(~) ()"
+    transferType <- if writable || constructOnly
+                    then propTransferType (propType prop)
+                    else return "()"
     let allowedOps = (if writable
                       then ["'AttrSet", "'AttrConstruct"]
                       else [])
@@ -363,16 +429,23 @@ genOneProperty owner prop = do
     indent $ do
             line $ "type AttrAllowedOps " <> it
                      <> " = '[ " <> T.intercalate ", " allowedOps <> "]"
+            line $ "type AttrBaseTypeConstraint " <> it <> " = " <> cls
             line $ "type AttrSetTypeConstraint " <> it
                      <> " = " <> inConstraint
-            line $ "type AttrBaseTypeConstraint " <> it <> " = " <> cls
+            line $ "type AttrTransferTypeConstraint " <> it
+                     <> " = " <> transferConstraint
+            line $ "type AttrTransferType " <> it <> " = " <> transferType
             line $ "type AttrGetType " <> it <> " = " <> outType
             line $ "type AttrLabel " <> it <> " = \"" <> propName prop <> "\""
             line $ "type AttrOrigin " <> it <> " = " <> name
-            line $ "attrGet _ = " <> getter
-            line $ "attrSet _ = " <> setter
-            line $ "attrConstruct _ = " <> constructor
-            line $ "attrClear _ = " <> clear
+            line $ "attrGet = " <> getter
+            line $ "attrSet = " <> setter
+            if writable || constructOnly
+              then do line $ "attrTransfer _ v = do"
+                      indent $ genPropTransfer "v" (propType prop)
+              else line $ "attrTransfer _ = undefined"
+            line $ "attrConstruct = " <> constructor
+            line $ "attrClear = " <> clear
 
 -- | Generate a placeholder property for those cases in which code
 -- generation failed.
@@ -383,11 +456,13 @@ genPlaceholderProperty owner prop = do
   let cName = (hyphensToCamelCase . propName) prop
       docSection = NamedSubsection PropertySection (lcFirst cName)
   export docSection it
-  line $ "data " <> it
+  bline $ "data " <> it
   line $ "instance AttrInfo " <> it <> " where"
   indent $ do
     line $ "type AttrAllowedOps " <> it <> " = '[]"
     line $ "type AttrSetTypeConstraint " <> it <> " = (~) ()"
+    line $ "type AttrTransferTypeConstraint " <> it <> " = (~) ()"
+    line $ "type AttrTransferType " <> it <> " = ()"
     line $ "type AttrBaseTypeConstraint " <> it <> " = (~) ()"
     line $ "type AttrGetType " <> it <> " = ()"
     line $ "type AttrLabel " <> it <> " = \"\""
@@ -396,6 +471,7 @@ genPlaceholderProperty owner prop = do
     line $ "attrSet = undefined"
     line $ "attrConstruct = undefined"
     line $ "attrClear = undefined"
+    line $ "attrTransfer = undefined"
 
 genProperties :: Name -> [Property] -> [Text] -> CodeGen ()
 genProperties n ownedProps allProps = do
@@ -405,7 +481,8 @@ genProperties n ownedProps allProps = do
       handleCGExc (\err -> do
                      line $ "-- XXX Generation of property \""
                               <> propName prop <> "\" of object \""
-                              <> name <> "\" failed: " <> describeCGError err
+                              <> name <> "\" failed."
+                     printCGError err
                      cppIf CPPOverloading (genPlaceholderProperty n prop))
                   (genOneProperty n prop)
 
